#!/usr/bin/env python
# coding=utf-8
# Copyright 2009 Facebook
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""A lightweight wrapper around MySQLdb.
Originally part of the Tornado framework.  The tornado.database module
is slated for removal in Tornado 3.0, and it is now available separately
as torndb.
"""

import copy
import os
import time

# from consumer.src import settings
try:
    import MySQLdb.constants
    import MySQLdb.converters
    import MySQLdb.cursors
except ImportError:
    try:
        import pymysql as MySQLdb

        MySQLdb.install_as_MySQLdb()
    except ImportError:
        # If MySQLdb isn't available this module won't actually be useable,
        # but we want it to at least be importable on readthedocs.org,
        # which has limitations on third-party modules.
        if 'READTHEDOCS' in os.environ:
            MySQLdb = None
        else:
            raise
from .db_base import DbBase

version = "0.3"
version_info = (0, 3, 0, 0)


class Connection(DbBase):
    """A lightweight wrapper around MySQLdb DB-API connections.
    The main value we provide is wrapping rows in a dict/object so that
    columns can be accessed by name. Typical usage::
        db = torndb.Connection("localhost", "mydatabase")
        for article in db.query("SELECT * FROM articles"):
            print article.title
    Cursors are hidden by the implementation, but other than that, the methods
    are very similar to the DB-API.
    We explicitly set the timezone to UTC and assume the character encoding to
    UTF-8 (can be changed) on all connections to avoid time zone and encoding errors.
    The sql_mode parameter is set by default to "traditional", which "gives an error instead of a warning"
    (http://dev.mysql.com/doc/refman/5.0/en/server-sql-mode.html). However, it can be set to
    any other mode including blank (None) thereby explicitly clearing the SQL mode.
    Arguments read_timeout and write_timeout can be passed using kwargs, if
    MySQLdb version >= 1.2.5 and MySQL version > 5.1.12.
    """

    def __init__(self, **kwargs):
        super(Connection, self).__init__(**kwargs)
        self.__init_base_args()
        self._db = None
        self._last_use_time = time.time()
        try:
            self.reconnect()
        except Exception:
            self._log.error("Cannot connect to MySQL on %s", self._host, exc_info=True)

    def __init_base_args(self):
        """"""
        self._host = self._kwargs.pop('host', '0.0.0.0')
        self._port = int(self._kwargs.pop('port', 3306))
        self._user = self._kwargs.pop('username', 'root')
        self._passwd = self._kwargs.pop('password', 'root')
        self._db = str(self._kwargs.pop('db_name', 'siterec_dashboard'))
        self._charset = self._kwargs.pop('charset', 'utf8')
        self._autocommit = self._kwargs.pop('autocommit', False)
        self._max_idle_time = float(self._kwargs.pop('max_idle_time', 7 * 3600))   # 保持重新连接时间
        self._connect_timeout = self._kwargs.pop('connect_timeout', 10)
        self._time_zone = self._kwargs.pop('time_zone', "+0:00")    # 时区
        self._sql_mode = self._kwargs.pop('sql_mode', None)
        self._use_unicode = self._kwargs.pop('use_unicode', True)   # 默认使用Unicode字符
        self._new_kwargs = {
            "user": self._user,
            "passwd": self._passwd,
            "db": self._db,
            "charset": self._charset,
            "connect_timeout": self._connect_timeout,
            "init_command": 'SET time_zone = "{}"'.format(self._time_zone),
            "sql_mode": self._sql_mode,
            "use_unicode": self._use_unicode,
        }

        # We accept a path to a MySQL socket file or a host(:port) string
        if "/" in self._host:
            self._new_kwargs["unix_socket"] = self._host
        else:
            self.socket = None
            pair = self._host.split(":")
            if len(pair) == 2:
                self._new_kwargs["host"] = pair[0]
                self._new_kwargs["port"] = int(pair[1])
            else:
                self._new_kwargs["host"] = self._host
                self._new_kwargs["port"] = self._port
        self._db_args = {**self._kwargs, **self._new_kwargs}

    def __del__(self):
        self.close()

    def rollback(self):
        self._db.rollback()

    def close(self):
        """Closes this database connection."""
        if getattr(self, "_db", None) is not None:
            self._db.close()
            self._db = None

    def reconnect(self):
        """Closes the existing database connection and re-opens it."""
        self.close()
        self._db = MySQLdb.connect(**self._db_args)
        self._db.autocommit(True)

    def iter(self, query, *parameters, **kwparameters):
        """Returns an iterator for the given query and parameters."""
        self._ensure_connected()
        cursor = MySQLdb.cursors.SSCursor(self._db)
        try:
            self._execute(cursor, query, parameters, kwparameters)
            column_names = [d[0] for d in cursor.description]
            for row in cursor:
                yield Row(list(zip(column_names, row)))
        finally:
            cursor.close()

    def query(self, query, *parameters, **kwparameters):
        """Returns a row list for the given query and parameters."""
        cursor = self._cursor()
        try:
            self._execute(cursor, query, parameters, kwparameters)
            column_names = [d[0] for d in cursor.description]
            return [list(zip(column_names, row)) for row in cursor]
        finally:
            cursor.close()

    def get(self, query, *parameters, **kwparameters):
        """Returns the (singular) row returned by the given query.
        If the query has no results, returns None.  If it has
        more than one result, raises an exception.
        """
        rows = self.query(query, *parameters, **kwparameters)
        if not rows:
            return None
        elif len(rows) > 1:
            raise TypeError("Multiple rows returned for Database.get() query")
        else:
            return rows[0]

    # rowcount is a more reasonable default return value than lastrowid,
    # but for historical compatibility execute() must return lastrowid.
    def execute(self, query, *parameters, **kwparameters):
        """Executes the given query, returning the lastrowid from the query."""
        return self.execute_lastrowid(query, *parameters, **kwparameters)

    def execute_lastrowid(self, query, *parameters, **kwparameters):
        """Executes the given query, returning the lastrowid from the query."""
        cursor = self._cursor()
        try:
            self._execute(cursor, query, parameters, kwparameters)
            return cursor.lastrowid
        finally:
            cursor.close()

    def execute_rowcount(self, query, *parameters, **kwparameters):
        """Executes the given query, returning the rowcount from the query."""
        cursor = self._cursor()
        try:
            self._execute(cursor, query, parameters, kwparameters)
            return cursor.rowcount
        finally:
            cursor.close()

    def executemany(self, query, parameters):
        """Executes the given query against all the given param sequences.
        We return the lastrowid from the query.
        """
        return self.executemany_lastrowid(query, parameters)

    def executemany_lastrowid(self, query, parameters):
        """Executes the given query against all the given param sequences.
        We return the lastrowid from the query.
        """
        cursor = self._cursor()
        try:
            cursor.executemany(query, parameters)
            return cursor.lastrowid
        finally:
            cursor.close()

    def executemany_rowcount(self, query, parameters):
        """Executes the given query against all the given param sequences.
        We return the rowcount from the query.
        """
        cursor = self._cursor()
        try:
            cursor.executemany(query, parameters)
            return cursor.rowcount
        finally:
            cursor.close()

    # -------------- 别名方法 --------------
    select = get
    selectmany = query

    update = delete = execute_rowcount
    updatemany = executemany_rowcount

    insert = execute_lastrowid
    insertmany = executemany_lastrowid

    def _ensure_connected(self):
        # Mysql by default closes client connections that are idle for
        # 8 hours, but the client library does not report this fact until
        # you try to perform a query and it fails.  Protect against this
        # case by preemptively closing and reopening the connection
        # if it has been idle for too long (7 hours by default).
        if (self._db is None or
                (time.time() - self._last_use_time > self._max_idle_time)):
            self.reconnect()
        self._last_use_time = time.time()

    def _cursor(self):
        self._ensure_connected()
        return self._db.cursor()

    def _execute(self, cursor, query, parameters, kwparameters):
        try:
            return cursor.execute(query, kwparameters or parameters)
        except Exception as e:
            self._log.error(e)
            raise


class Row(dict):
    """A dict that allows for object-like property access syntax."""

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)


if MySQLdb is not None:

    if MySQLdb.__package__ != 'pymysql':
        # Fix the access conversions to properly recognize unicode/binary
        FIELD_TYPE = MySQLdb.constants.FIELD_TYPE
        FLAG = MySQLdb.constants.FLAG
        CONVERSIONS = copy.copy(MySQLdb.converters.conversions)

        field_types = [FIELD_TYPE.BLOB, FIELD_TYPE.STRING, FIELD_TYPE.VAR_STRING]
        if 'VARCHAR' in vars(FIELD_TYPE):
            field_types.append(FIELD_TYPE.VARCHAR)

        # for field_type in field_types:
        #     CONVERSIONS[field_type] = [(FLAG.BINARY, str)] + CONVERSIONS[field_type]
    else:
        CONVERSIONS = {}

    # Alias some common MySQL exceptions
    IntegrityError = MySQLdb.IntegrityError
    OperationalError = MySQLdb.OperationalError
